@@ -21,8 +21,16 @@
  */
 #if PG_VERSION_NUM < 90300
 #error "OGR FDW requires PostgreSQL version 9.3 or higher"
+
 #else
 
+/*
+ * Definition of stringToQualifiedNameList
+ */
+#if PG_VERSION_NUM >= 100000
+#include "utils/regproc.h"
+#endif
+
 /*
  * Local structures
  */
@@ -156,34 +164,35 @@ static void ogrReadColumnData(OgrFdwState *state);
 /* Global to hold GEOMETRYOID */
 Oid GEOMETRYOID = InvalidOid;
 
-
+#if GDAL_VERSION_MAJOR >= 2 && GDAL_VERSION_MINOR >= 1
+static void
+ogrErrorHandler(CPLErr eErrClass, int err_no, const char *msg)
+{
+	switch (eErrClass)
+	{
+		case CE_None:
+			elog(NOTICE, "[%d] %s", err_no, msg);
+			break;
+		case CE_Debug:
+			elog(DEBUG2, "[%d] %s", err_no, msg);
+			break;
+		case CE_Warning:
+			elog(WARNING, "[%d] %s", err_no, msg);
+			break;
+		case CE_Failure:
+		case CE_Fatal:
+		default:
+			elog(ERROR, "[%d] %s", err_no, msg);
+			break;
+	}
+	return;
+}
+#endif
 
 void
 _PG_init(void)
 {
-	// DefineCustomIntVariable("mysql_fdw.wait_timeout",
-	// 						"Server-side wait_timeout",
-	// 						"Set the maximum wait_timeout"
-	// 						"use to set the MySQL session timeout",
-	// 						&wait_timeout,
-	// 						WAIT_TIMEOUT,
-	// 						0,
-	// 						INT_MAX,
-	// 						PGC_USERSET,
-	// 						0,
-	// 						NULL,
-	// 						NULL,
-	// 						NULL);
-
-	/*
-	 * We assume PostGIS is installed in 'public' and if we cannot
-	 * find it, we'll treat all geometry from OGR as bytea.
-	 */
-	// const char *typname = "geometry";
-	// Oid namesp = LookupExplicitNamespace("public", false);
-	// Oid typoid = GetSysCacheOid2(TYPENAMENSP, CStringGetDatum(typname), ObjectIdGetDatum(namesp));
 	Oid typoid = TypenameGetTypid("geometry");
-
 	if (OidIsValid(typoid) && get_typisdefined(typoid))
 	{
 		GEOMETRYOID = typoid;
@@ -194,6 +203,12 @@ _PG_init(void)
 	}
 
 	on_proc_exit(&ogr_fdw_exit, PointerGetDatum(NULL));
+
+#if GDAL_VERSION_MAJOR >= 2 && GDAL_VERSION_MINOR >= 1
+	/* Hook up the GDAL error handlers to PgSQL elog() */
+	CPLSetErrorHandler(ogrErrorHandler);
+	CPLSetCurrentErrorHandlerCatchDebug(true);
+#endif
 }
 
 /*
@@ -223,7 +238,7 @@ ogr_fdw_handler(PG_FUNCTION_ARGS)
 	fdwroutine->IterateForeignScan = ogrIterateForeignScan;
 	fdwroutine->ReScanForeignScan = ogrReScanForeignScan;
 	fdwroutine->EndForeignScan = ogrEndForeignScan;
-	
+
 	/* Write support */
 	fdwroutine->AddForeignUpdateTargets = ogrAddForeignUpdateTargets;
 	fdwroutine->BeginForeignModify = ogrBeginForeignModify;
@@ -247,7 +262,7 @@ ogr_fdw_handler(PG_FUNCTION_ARGS)
  * and in FDW options validation.
  */
 static GDALDatasetH
-ogrGetDataSource(const char *source, const char *driver, bool updateable, 
+ogrGetDataSource(const char *source, const char *driver, bool updateable,
                  const char *config_options, const char *open_options)
 {
 	GDALDatasetH ogr_ds = NULL;
@@ -255,7 +270,7 @@ ogrGetDataSource(const char *source, const char *driver, bool updateable,
 	char **open_option_list = NULL;
 #if GDAL_VERSION_MAJOR >= 2
 	unsigned int open_flags = GDAL_OF_VECTOR;
-	
+
 	if ( updateable )
 		open_flags |= GDAL_OF_UPDATE;
 	else
@@ -274,7 +289,7 @@ ogrGetDataSource(const char *source, const char *driver, bool updateable,
 			value = CPLParseNameValue(*option_iter, &key);
 			if ( ! (key && value) )
 				elog(ERROR, "bad config option string '%s'", config_options);
-			
+
 			elog(DEBUG1, "GDAL config option '%s' set to '%s'", key, value);
 			CPLSetConfigOption(key, value);
 			CPLFree(key);
@@ -320,9 +335,9 @@ ogrGetDataSource(const char *source, const char *driver, bool updateable,
 #if GDAL_VERSION_MAJOR < 2
 		ogr_ds = OGROpen(source, updateable, &ogr_dr);
 #else
-		ogr_ds = GDALOpenEx(source, 
-		                    open_flags, 
-		                    NULL, 
+		ogr_ds = GDALOpenEx(source,
+		                    open_flags,
+		                    NULL,
 		                    (const char *const *)open_option_list,
 		                    NULL);
 #endif
@@ -346,7 +361,7 @@ ogrGetDataSource(const char *source, const char *driver, bool updateable,
  				 errmsg("unable to connect to data source \"%s\"", source)));
 		}
 	}
-	
+
 	CSLDestroy( open_option_list );
 
 	return ogr_ds;
@@ -367,7 +382,7 @@ ogrCanReallyCountFast(const OgrConnection *con)
 	return false;
 }
 
-static void 
+static void
 ogrEreportError(const char *errstr)
 {
 	const char *ogrerr = CPLGetLastErrorMsg();
@@ -413,7 +428,7 @@ ogrGetConnectionFromServer(Oid foreignserverid, bool updateable)
 	memset(&ogr, 0, sizeof(OgrConnection));
 	ogr.ds_updateable = OGR_UPDATEABLE_UNSET;
 	ogr.lyr_updateable = OGR_UPDATEABLE_UNSET;
-	
+
 	server = GetForeignServer(foreignserverid);
 
 	foreach(cell, server->options)
@@ -432,7 +447,7 @@ ogrGetConnectionFromServer(Oid foreignserverid, bool updateable)
 			if ( defGetBoolean(def) )
 				ogr.ds_updateable = OGR_UPDATEABLE_TRUE;
 			else
-				ogr.ds_updateable = OGR_UPDATEABLE_FALSE;				
+				ogr.ds_updateable = OGR_UPDATEABLE_FALSE;
 		}
 	}
 
@@ -444,7 +459,7 @@ ogrGetConnectionFromServer(Oid foreignserverid, bool updateable)
 			(errcode(ERRCODE_FDW_ERROR),
 			 errmsg("updates are not allowed on foreign server '%s'", server->servername),
 			 errhint("ALTER FOREIGN SERVER %s OPTIONS (SET updatable 'true')", server->servername)));
-		
+
 	/*
 	 * TODO: Connections happen twice for each query, having a
 	 * connection pool will certainly make things faster.
@@ -488,7 +503,7 @@ ogrGetConnectionFromTable(Oid foreigntableid, bool updateable)
 			if ( defGetBoolean(def) )
 				ogr.lyr_updateable = OGR_UPDATEABLE_TRUE;
 			else
-				ogr.lyr_updateable = OGR_UPDATEABLE_FALSE;				
+				ogr.lyr_updateable = OGR_UPDATEABLE_FALSE;
 		}
 	}
 
@@ -514,6 +529,7 @@ ogrGetConnectionFromTable(Oid foreigntableid, bool updateable)
 				: errhint("Does the layer exist?")
 				));
 	}
+	ogr.lyr_utf8 = OGR_L_TestCapability(ogr.lyr, OLCStringsAsUTF8);
 
 	return ogr;
 }
@@ -641,8 +657,8 @@ getOgrFdwState(Oid foreigntableid, OgrFdwStateType state_type)
 	OgrFdwState *state;
 	size_t size;
 	bool updateable = false;
-	
-	switch (state_type) 
+
+	switch (state_type)
 	{
 		case OGR_PLAN_STATE:
 			size = sizeof(OgrFdwPlanState);
@@ -659,7 +675,7 @@ getOgrFdwState(Oid foreigntableid, OgrFdwStateType state_type)
 		default:
 			elog(ERROR, "invalid state type");
 	}
-	
+
 	state = palloc0(size);
 	state->type = state_type;
 
@@ -862,7 +878,7 @@ ogrGetForeignPlan(PlannerInfo *root,
 
 static void
 pgCanConvertToOgr(Oid pg_type, OGRFieldType ogr_type, const char *colname, const char *tblname)
-{	
+{
 	if ( pg_type == BOOLOID && ogr_type == OFTInteger )
 		return;
 	else if ( pg_type == INT2OID && ogr_type == OFTInteger )
@@ -872,13 +888,13 @@ pgCanConvertToOgr(Oid pg_type, OGRFieldType ogr_type, const char *colname, const
 	else if ( pg_type == INT8OID )
 	{
 #if GDAL_VERSION_MAJOR >= 2
-		if ( ogr_type == OFTInteger64 ) 
+		if ( ogr_type == OFTInteger64 )
 			return;
 #else
 		if ( ogr_type == OFTInteger )
 			return;
 #endif
-	} 
+	}
 	else if ( pg_type == NUMERICOID && ogr_type == OFTReal )
 		return;
 	else if ( pg_type == FLOAT4OID && ogr_type == OFTReal )
@@ -903,15 +919,13 @@ pgCanConvertToOgr(Oid pg_type, OGRFieldType ogr_type, const char *colname, const
 		return;
 	else if ( pg_type == TIMESTAMPOID && ogr_type == OFTDateTime )
 		return;
-	else if ( pg_type == TIMESTAMPOID && ogr_type == OFTDateTime )
-		return;
-	
+
 	ereport(ERROR, (
 			errcode(ERRCODE_FDW_INVALID_DATA_TYPE),
 			errmsg("column \"%s\" of foreign table \"%s\" converts \"%s\" to OGR \"%s\"",
 				colname, tblname,
 				format_type_be(pg_type), OGR_GetFieldTypeName(ogr_type))
-			));	
+			));
 }
 
 static void
@@ -989,7 +1003,7 @@ ogrCanConvertToPg(OGRFieldType ogr_type, Oid pg_type, const char *colname, const
 
 static char *hexchr = "0123456789ABCDEF";
 
-static char * 
+static char *
 ogrBytesToHex(unsigned char *bytes, size_t size)
 {
 	char *hex;
@@ -1085,7 +1099,7 @@ ogrReadColumnData(OgrFdwState *state)
 	/* Get OGR metadata ready */
 	dfn = OGR_L_GetLayerDefn(state->ogr.lyr);
 	ogr_ncols = OGR_FD_GetFieldCount(dfn);
-#if GDAL_VERSION_MAJOR >= 2 || GDAL_VERSION_MINOR >= 11
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(1,11,0)
 	ogr_geom_count = OGR_FD_GetGeomFieldCount(dfn);
 #else
 	ogr_geom_count = ( OGR_FD_GetGeomType(dfn) != wkbNone ) ? 1 : 0;
@@ -1118,8 +1132,12 @@ ogrReadColumnData(OgrFdwState *state)
 		ListCell *lc;
 		OgrFieldEntry *found_entry;
 		OgrFieldEntry entry;
-		
+
+#if PG_VERSION_NUM >= 110000
+		Form_pg_attribute att_tuple = &tupdesc->attrs[i];
+#else
 		Form_pg_attribute att_tuple = tupdesc->attrs[i];
+#endif
 		OgrFdwColumn col = tbl->cols[i];
 		col.pgattnum = att_tuple->attnum;
 		col.pgtype = att_tuple->atttypid;
@@ -1135,7 +1153,7 @@ ogrReadColumnData(OgrFdwState *state)
 		getTypeBinaryInputInfo(col.pgtype, &col.pgrecvfunc, &col.pgrecvioparam);
 		getTypeOutputInfo(col.pgtype, &col.pgoutputfunc, &col.pgoutputvarlena);
 		getTypeBinaryOutputInfo(col.pgtype, &col.pgsendfunc, &col.pgsendvarlena);
-			
+
 		/* Get the PgSQL column name */
 		col.pgname = get_relid_attribute_name(rel->rd_id, att_tuple->attnum);
 
@@ -1150,7 +1168,7 @@ ogrReadColumnData(OgrFdwState *state)
 			tbl->cols[i] = col;
 			continue;
 		}
-		
+
 		/* If the OGR source has geometries, can we match them to Pg columns? */
 		/* We'll match to the first ones we find, irrespective of name */
 		if ( geom_count < ogr_geom_count && col.pgtype == GEOMETRYOID )
@@ -1211,7 +1229,7 @@ ogrReadColumnData(OgrFdwState *state)
 	elog(DEBUG2, "ogrReadColumnData matched %d FID, %d GEOM, %d FIELDS out of %d PGSQL COLUMNS", fid_count, geom_count, field_count, tbl->ncols);
 
 	/* Clean up */
-	
+
 	state->table = tbl;
 	for( i = 0; i < 2*ogr_ncols; i++ )
 		if ( ogr_fields[i].fldname ) pfree(ogr_fields[i].fldname);
@@ -1221,22 +1239,22 @@ ogrReadColumnData(OgrFdwState *state)
 	return;
 }
 
-/* 
+/*
  * ogrLookupGeometryFunctionOid
- * 
- * Find the procedure Oids of useful functions so we can call 
+ *
+ * Find the procedure Oids of useful functions so we can call
  * them later.
  */
-static Oid 
+static Oid
 ogrLookupGeometryFunctionOid(const char *proname)
 {
 	List *names;
 	FuncCandidateList clist;
-	
+
 	/* This only works if PostGIS is installed */
 	if ( GEOMETRYOID == InvalidOid || GEOMETRYOID == BYTEAOID )
 		return InvalidOid;
-	
+
 	names = stringToQualifiedNameList(proname);
 #if PG_VERSION_NUM < 90400
 	clist = FuncnameGetCandidates(names, -1, NIL, false, false);
@@ -1260,7 +1278,7 @@ ogrLookupGeometryFunctionOid(const char *proname)
 	{
 		return clist->oid;
 	}
-	
+
 	return InvalidOid;
 }
 
@@ -1279,7 +1297,7 @@ ogrBeginForeignScan(ForeignScanState *node, int eflags)
 
 	/* Read the OGR layer definition and PgSQL foreign table definitions */
 	ogrReadColumnData(state);
-	
+
 	/* Collect the procedure Oids for PostGIS functions we might need */
 	execstate->setsridfunc = ogrLookupGeometryFunctionOid("st_setsrid");
 	execstate->typmodsridfunc = ogrLookupGeometryFunctionOid("postgis_typmod_srid");
@@ -1294,7 +1312,7 @@ ogrBeginForeignScan(ForeignScanState *node, int eflags)
 		if ( err != OGRERR_NONE )
 		{
 			const char *ogrerr = CPLGetLastErrorMsg();
-			
+
 			if ( ogrerr && ! streq(ogrerr,"") )
 			{
 				ereport(NOTICE,
@@ -1325,7 +1343,7 @@ ogrBeginForeignScan(ForeignScanState *node, int eflags)
  * Rather than explicitly try and form PgSQL datums, use the type
  * input functions, that accept cstring representations, and convert
  * to the input format. We have to lookup the right input function for
- * each column in the foreign table. 
+ * each column in the foreign table.
  */
 static Datum
 pgDatumFromCString(const char *cstr, Oid pgtype, int pgtypmod, Oid pginputfunc)
@@ -1333,19 +1351,9 @@ pgDatumFromCString(const char *cstr, Oid pgtype, int pgtypmod, Oid pginputfunc)
 	Datum value;
 	Datum cdata = CStringGetDatum(cstr);
 
-	/* pgtypmod will be -1 for types w/o typmod  */
-	// if ( pgtypmod >= 0 )
-	// {
-		/* These functions require a type modifier */
-		value = OidFunctionCall3(pginputfunc, cdata,
-			ObjectIdGetDatum(InvalidOid),
-			Int32GetDatum(pgtypmod));
-	// }
-	// else
-	// {
-	// 	/* These functions don't */
-	// 	value = OidFunctionCall1(pginputfunc, cdata);
-	// }
+	value = OidFunctionCall3(pginputfunc, cdata,
+		ObjectIdGetDatum(InvalidOid),
+		Int32GetDatum(pgtypmod));
 
 	return value;
 }
@@ -1434,7 +1442,7 @@ ogrFeatureToSlot(const OGRFeatureH feat, TupleTableSlot *slot, const OgrFdwExecS
 			unsigned char *wkb;
 			OGRErr err;
 
-#if GDAL_VERSION_MAJOR >= 2 || GDAL_VERSION_MINOR >= 11
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(1,11,0)
 			OGRGeometryH geom = OGR_F_GetGeomFieldRef(feat, ogrfldnum);
 #else
 			OGRGeometryH geom = OGR_F_GetGeometryRef(feat);
@@ -1514,35 +1522,41 @@ ogrFeatureToSlot(const OGRFeatureH feat, TupleTableSlot *slot, const OgrFdwExecS
 				nulls[i] = false;
 				values[i] = OidFunctionCall1(col.pgrecvfunc, PointerGetDatum(&strinfo));
 #endif
-				
-				/* 
+
+				/*
 				 * Apply the typmod restriction to the incoming geometry, so it's
 				 * not really a restriction anymore, it's more like a requirement.
-				 * 
+				 *
 				 * TODO: In the case where the OGR input actually *knows* what SRID
 				 * it is, we should actually apply *that* and let the restriction run
 				 * its usual course.
 				 */
-				if ( have_typmod_funcs && col.pgtypmod >= 0 ) 
+				if ( have_typmod_funcs && col.pgtypmod >= 0 )
 				{
 					Datum srid = OidFunctionCall1(execstate->typmodsridfunc, Int32GetDatum(col.pgtypmod));
 					values[i] = OidFunctionCall2(execstate->setsridfunc, values[i], srid);
-				}			
+				}
 			}
 			else
 			{
 				elog(NOTICE, "conversion to geometry called with column type not equal to bytea or geometry");
 				ogrNullSlot(values, nulls, i);
 			}
-			
+
 		}
 		else if ( ogrvariant == OGR_FIELD )
 		{
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(2,2,0)
+			int field_not_null = OGR_F_IsFieldSet(feat, ogrfldnum) && ! OGR_F_IsFieldNull(feat, ogrfldnum);
+#else
+			int field_not_null = OGR_F_IsFieldSet(feat, ogrfldnum);
+#endif
+
 			/* Ensure that the OGR data type fits the destination Pg column */
 			ogrCanConvertToPg(ogrfldtype, pgtype, pgname, tbl->tblname);
 
 			/* Only convert non-null fields */
-			if ( OGR_F_IsFieldSet(feat, ogrfldnum) )
+			if ( field_not_null )
 			{
 				switch(ogrfldtype)
 				{
@@ -1573,11 +1587,17 @@ ogrFeatureToSlot(const OGRFeatureH feat, TupleTableSlot *slot, const OgrFdwExecS
 						 * Handling numbers directly would be faster, but require a lot of extra code.
 						 * For now, we go via text.
 						 */
-						const char *cstr = OGR_F_GetFieldAsString(feat, ogrfldnum);
-						if ( cstr )
+						const char *cstr_in = OGR_F_GetFieldAsString(feat, ogrfldnum);
+						size_t cstr_len = cstr_in ? strlen(cstr_in) : 0;
+						if ( cstr_in && cstr_len > 0 )
 						{
+							char *cstr_decoded;
+							if(execstate->ogr.lyr_utf8)
+								cstr_decoded = pg_any_to_server(cstr_in, cstr_len, PG_UTF8);
+							else
+								cstr_decoded = pstrdup(cstr_in);
 							nulls[i] = false;
-							values[i] = pgDatumFromCString(cstr, pgtype, pgtypmod, pginputfunc);
+							values[i] = pgDatumFromCString(cstr_decoded, pgtype, pgtypmod, pginputfunc);
 						}
 						else
 						{
@@ -1613,7 +1633,6 @@ ogrFeatureToSlot(const OGRFeatureH feat, TupleTableSlot *slot, const OgrFdwExecS
 						{
 							snprintf(cstr, 256, "%d-%02d-%02d %02d:%02d:%02d", year, month, day, hour, minute, second);
 						}
-
 						nulls[i] = false;
 						values[i] = pgDatumFromCString(cstr, pgtype, pgtypmod, pginputfunc);
 						break;
@@ -1682,18 +1701,18 @@ ogrEwkbStripSrid(unsigned char *wkb, size_t wkbsize)
 	/* has_z = type & 0x80000000; */
 	/* has_m = type & 0x40000000; */
 	has_srid = type & 0x20000000;
-	
+
 	/* Flatten SRID flag away */
 	type &= 0xDFFFFFFF;
 	memcpy(wkb+1, &type, 4);
-	
+
 	/* If there was an SRID number embedded, overwrite it */
 	if ( has_srid )
 	{
 		newwkbsize -= 4; /* no space for SRID number needed */
 		memmove(wkb+5, wkb+9, newwkbsize - 5);
 	}
-	
+
 	return newwkbsize;
 }
 
@@ -1706,7 +1725,7 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 	TupleDesc tupdesc = slot->tts_tupleDescriptor;
 
 	int year, month, day, hour, minute, second;
-	
+
 	/* Prepare date-time part tokens for use later */
 	char txtyear[STR_MAX_LEN];
 	char txtmonth[STR_MAX_LEN];
@@ -1768,17 +1787,17 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 					elog(ERROR, "unable to handle non-integer fid");
 				}
 			}
-			continue;				
+			continue;
 		}
-		
+
 		/* TODO: For updates, we should only set the fields that are */
 		/*       in the target list, and flag the others as unchanged */
 		if ( ogrvariant == OGR_GEOMETRY )
-		{			
+		{
 			OGRErr err;
 			if ( nulls[i] )
 			{
-#if GDAL_VERSION_MAJOR >= 2 || GDAL_VERSION_MINOR >= 11
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(1,11,0)
 				err = OGR_F_SetGeomFieldDirectly(feat, ogrfldnum, NULL);
 #else
 				err = OGR_F_SetGeometryDirectly(feat, NULL);
@@ -1792,14 +1811,14 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 				unsigned char *wkb = (unsigned char *)VARDATA(wkb_bytea);
 				int wkbsize = VARSIZE(wkb_bytea) - VARHDRSZ;
 				wkbsize = ogrEwkbStripSrid(wkb, wkbsize);
-				
+
 				/* TODO, create geometry with SRS of table? */
 				err = OGR_G_CreateFromWkb(wkb, NULL, &geom, wkbsize);
 				if ( wkb_bytea ) pfree(wkb_bytea);
-				if ( err != OGRERR_NONE ) 
+				if ( err != OGRERR_NONE )
 					return err;
-				
-#if GDAL_VERSION_MAJOR >= 2 || GDAL_VERSION_MINOR >= 11
+
+#if GDAL_VERSION_NUM >= GDAL_COMPUTE_VERSION(1,11,0)
 				err = OGR_F_SetGeomFieldDirectly(feat, ogrfldnum, geom);
 #else
 				err = OGR_F_SetGeometryDirectly(feat, geom);
@@ -1841,7 +1860,7 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 				case INT8OID:
 				{
 					int64 val = DatumGetInt64(values[i]);
-#if GDAL_VERSION_MAJOR >= 2					
+#if GDAL_VERSION_MAJOR >= 2
 					OGR_F_SetFieldInteger64(feat, ogrfldnum, val);
 #else
 					if ( val < INT_MAX )
@@ -1850,28 +1869,28 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 						elog(ERROR, "unable to coerce int64 into int32 OGR field");
 #endif
 					break;
-					
+
 				}
-				
+
 				case NUMERICOID:
 				{
 					Datum d;
 					float8 f;
-					
+
 					/* Convert to string */
 					d = OidFunctionCall1(pgoutputfunc, values[i]);
 					/* Convert back to float8 */
 					f = DatumGetFloat8(DirectFunctionCall1(float8in, d));
 
 					OGR_F_SetFieldDouble(feat, ogrfldnum, f);
-					break;					
+					break;
 				}
-				case FLOAT4OID:	
+				case FLOAT4OID:
 				{
 					OGR_F_SetFieldDouble(feat, ogrfldnum, DatumGetFloat4(values[i]));
 					break;
 				}
-				case FLOAT8OID:	
+				case FLOAT8OID:
 				{
 					OGR_F_SetFieldDouble(feat, ogrfldnum, DatumGetFloat8(values[i]));
 					break;
@@ -1890,7 +1909,7 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 					pfree(str);
 					break;
 				}
-				
+
 				case CHAROID: /* char */
 				{
 					char str[2];
@@ -1899,20 +1918,20 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 					OGR_F_SetFieldString(feat, ogrfldnum, str);
 					break;
 				}
-				
+
 				case BYTEAOID:
 				{
 					bytea *varlena = PG_DETOAST_DATUM(values[i]);
 					size_t varsize = VARSIZE(varlena) - VARHDRSZ;
 					OGR_F_SetFieldBinary(feat, ogrfldnum, varsize, (GByte *)VARDATA(varlena));
-					break;				
+					break;
 				}
 
 				case DATEOID:
 				{
 					/* Convert date to timestamp */
 					Datum d = DirectFunctionCall1(date_timestamp, values[i]);
-					
+
 					/* Read out the parts */
 					year = lround(DatumGetFloat8(DirectFunctionCall2(timestamp_part, PointerGetDatum(txtyear), d)));
 					month = lround(DatumGetFloat8(DirectFunctionCall2(timestamp_part, PointerGetDatum(txtmonth), d)));
@@ -1920,7 +1939,7 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 					OGR_F_SetFieldDateTime(feat, ogrfldnum, year, month, day, 0, 0, 0, 0);
 					break;
 				}
-				
+
 				/* TODO: handle time zones explicitly */
 				case TIMEOID:
 				case TIMETZOID:
@@ -1953,7 +1972,7 @@ ogrSlotToFeature(const TupleTableSlot *slot, OGRFeatureH feat, const OgrFdwTable
 				{
 					elog(ERROR, "OGR FDW unsupported PgSQL column type in \"%s\", %d", pgname, pgtype);
 					return OGRERR_FAILURE;
-				}				
+				}
 			}
 		}
 		/* Fill in unmatched columns with NULL */
@@ -2002,7 +2021,8 @@ ogrIterateForeignScan(ForeignScanState *node)
 	}
 
 	/* If we rectreive a feature from OGR, copy it over into the slot */
-	if ( (feat = OGR_L_GetNextFeature(execstate->ogr.lyr)) )
+	feat = OGR_L_GetNextFeature(execstate->ogr.lyr);
+	if ( feat )
 	{
 		/* convert result to arrays of values and null indicators */
 		if ( OGRERR_NONE != ogrFeatureToSlot(feat, slot, execstate) )
@@ -2058,14 +2078,14 @@ ogrEndForeignScan(ForeignScanState *node)
 
 // OgrFdwTable *tbl;
 
-/* if the scanning functions above respected the targetlist, 
+/* if the scanning functions above respected the targetlist,
 we would only be getting back the SET target=foo columns in the slots below,
 so we would need to add the "fid" to all targetlists (and also disallow fid changing
 perhaps).
 
 since we always pull complete tables in the scan functions, the
 slots below are basically full tables, in fact they include (?) one entry
-for each OGR column, even when the table does not include the column, 
+for each OGR column, even when the table does not include the column,
 just nulling out the entries that are not in the table definition
 
 it might be better to update the scan code to properly manage target lists
@@ -2084,9 +2104,14 @@ static int ogrGetFidColumn(const TupleDesc td)
 	int i;
 	for ( i = 0; i < td->natts; i++ )
 	{
+#if PG_VERSION_NUM >= 110000
+		NameData attname = td->attrs[i].attname;
+		Oid atttypeid = td->attrs[i].atttypid;
+#else
 		NameData attname = td->attrs[i]->attname;
 		Oid atttypeid = td->attrs[i]->atttypid;
-		if ( (atttypeid == INT4OID || atttypeid == INT8OID) && 
+#endif
+		if ( (atttypeid == INT4OID || atttypeid == INT8OID) &&
 		     strcaseeq("fid", attname.data) )
 		{
 			return i;
@@ -2097,7 +2122,7 @@ static int ogrGetFidColumn(const TupleDesc td)
 
 /*
  * ogrAddForeignUpdateTargets
- * 
+ *
  * For now we no-op this callback, as we are making the presence of
  * "fid" in the FDW table definition a requirement for any update.
  * It might be possible to add nonexisting "junk" columns? In which case
@@ -2114,14 +2139,17 @@ static void ogrAddForeignUpdateTargets (Query *parsetree,
 	TargetEntry *tle;
 	TupleDesc tupdesc = target_relation->rd_att;
 	int fid_column = ogrGetFidColumn(tupdesc);
-	
+
 	elog(DEBUG2, "ogrAddForeignUpdateTargets");
-	
+
 	if ( fid_column < 0 )
 		elog(ERROR,"table '%s' does not have a 'fid' column", RelationGetRelationName(target_relation));
-	
+
+#if PG_VERSION_NUM >= 110000
+	att = &tupdesc->attrs[fid_column];
+#else
 	att = tupdesc->attrs[fid_column];
-	
+#endif
 	/* Make a Var representing the desired value */
 	var = makeVar(parsetree->resultRelation,
 			att->attnum,
@@ -2137,18 +2165,18 @@ static void ogrAddForeignUpdateTargets (Query *parsetree,
 			true);
 
 	parsetree->targetList = lappend(parsetree->targetList, tle);
-	
+
 	foreach(cell, parsetree->targetList)
 	{
 		TargetEntry *target = (TargetEntry *) lfirst(cell);
 		elog(DEBUG4, "parsetree->targetList %s:%d", target->resname, target->resno);
 	}
-	
+
 	return;
 
 }
 
-/* 
+/*
  * ogrBeginForeignModify
  * For now the only thing we'll do here is set up the connection
  * and pass that on to the next functions.
@@ -2161,7 +2189,7 @@ static void ogrBeginForeignModify (ModifyTableState *mtstate,
 {
 	Oid foreigntableid;
 	OgrFdwState *state;
-	
+
 	elog(DEBUG2, "ogrBeginForeignModify");
 
 	foreigntableid = RelationGetRelid(rinfo->ri_RelationDesc);
@@ -2169,7 +2197,7 @@ static void ogrBeginForeignModify (ModifyTableState *mtstate,
 
 	/* Read the OGR layer definition and PgSQL foreign table definitions */
 	ogrReadColumnData(state);
-	
+
 	/* Save OGR connection, etc, for later */
 	rinfo->ri_FdwState = state;
 	return;
@@ -2177,7 +2205,7 @@ static void ogrBeginForeignModify (ModifyTableState *mtstate,
 
 /*
  * ogrExecForeignUpdate
- * Find out what the fid is, get the OGR feature for that FID, 
+ * Find out what the fid is, get the OGR feature for that FID,
  * and then update the values on that feature.
  */
 static TupleTableSlot *ogrExecForeignUpdate (EState *estate,
@@ -2195,23 +2223,26 @@ static TupleTableSlot *ogrExecForeignUpdate (EState *estate,
 	int64 fid;
 	OGRFeatureH feat;
 	OGRErr err;
-	
+
 	/* Is there a fid column? */
 	fid_column = ogrGetFidColumn(td);
 	if ( fid_column < 0 )
 		elog(ERROR, "cannot find 'fid' column in table '%s'", get_rel_name(foreigntableid));
-	
+
 	/* What is the value of the FID for this record? */
 	fid_datum = slot->tts_values[fid_column];
+#if PG_VERSION_NUM >= 110000
+	fid_type = td->attrs[fid_column].atttypid;
+#else
 	fid_type = td->attrs[fid_column]->atttypid;
-
+#endif
 	if ( fid_type == INT8OID )
 		fid = DatumGetInt64(fid_datum);
 	else
 		fid = DatumGetInt32(fid_datum);
 
 	elog(DEBUG2, "ogrExecForeignUpdate fid=" OGR_FDW_FRMT_INT64, OGR_FDW_CAST_INT64(fid));
-	
+
 	/* Get the OGR feature for this fid */
 	feat = OGR_L_GetFeature (modstate->ogr.lyr, fid);
 
@@ -2223,15 +2254,15 @@ static TupleTableSlot *ogrExecForeignUpdate (EState *estate,
 	err = ogrSlotToFeature(slot, feat, modstate->table);
 	if ( err != OGRERR_NONE )
 		ogrEreportError("failure populating OGR feature");
-	
+
 	err = OGR_L_SetFeature(modstate->ogr.lyr, feat);
 	if ( err != OGRERR_NONE )
 		ogrEreportError("failure writing back OGR feature");
-	
+
 	OGR_F_Destroy(feat);
-	
+
 	/* TODO: slot handling? what happens with RETURNING clauses? */
-	
+
 	return slot;
 }
 
@@ -2265,7 +2296,7 @@ static TupleTableSlot *ogrExecForeignUpdate (EState *estate,
 //     bool        tdhasoid;       /* tuple has oid attribute in its header */
 //     int         tdrefcount;     /* reference count, or -1 if not counting */
 // }   *TupleDesc;
-// 
+//
 
 // typedef struct ResultRelInfo
 // {
@@ -2350,7 +2381,7 @@ static TupleTableSlot *ogrExecForeignInsert (EState *estate,
 	GIntBig fid;
 
 	elog(DEBUG2, "ogrExecForeignInsert");
-	
+
 	/* Copy the data from the slot onto the feature */
 	if ( ! feat )
 		ogrEreportError("failure creating OGR feature");
@@ -2358,12 +2389,12 @@ static TupleTableSlot *ogrExecForeignInsert (EState *estate,
 	err = ogrSlotToFeature(slot, feat, modstate->table);
 	if ( err != OGRERR_NONE )
 		ogrEreportError("failure populating OGR feature");
-				
+
 	err = OGR_L_CreateFeature(modstate->ogr.lyr, feat);
 	if ( err != OGRERR_NONE )
 		ogrEreportError("failure writing OGR feature");
 
-	fid = OGR_F_GetFID(feat);	
+	fid = OGR_F_GetFID(feat);
 	OGR_F_Destroy(feat);
 
 	/* Update the FID for RETURNING slot */
@@ -2374,12 +2405,12 @@ static TupleTableSlot *ogrExecForeignInsert (EState *estate,
 		slot->tts_isnull[fid_column] = false;
 		slot->tts_nvalid++;
 	}
-	
+
 	return slot;
 }
-					
 
-					
+
+
 static TupleTableSlot *ogrExecForeignDelete (EState *estate,
 					ResultRelInfo *rinfo,
 					TupleTableSlot *slot,
@@ -2399,38 +2430,42 @@ static TupleTableSlot *ogrExecForeignDelete (EState *estate,
 	fid_column = ogrGetFidColumn(td);
 	if ( fid_column < 0 )
 		elog(ERROR, "cannot find 'fid' column in table '%s'", get_rel_name(foreigntableid));
-	
+
 	/* What is the value of the FID for this record? */
 	fid_datum = planSlot->tts_values[fid_column];
+#if PG_VERSION_NUM >= 110000
+	fid_type = td->attrs[fid_column].atttypid;
+#else
 	fid_type = td->attrs[fid_column]->atttypid;
+#endif
 
 	if ( fid_type == INT8OID )
 		fid = DatumGetInt64(fid_datum);
 	else
 		fid = DatumGetInt32(fid_datum);
-	
+
 	elog(DEBUG2, "ogrExecForeignDelete fid=" OGR_FDW_FRMT_INT64, OGR_FDW_CAST_INT64(fid));
-	
+
 	/* Delete the OGR feature for this fid */
 	err = OGR_L_DeleteFeature(modstate->ogr.lyr, fid);
-	
-	if ( err != OGRERR_NONE ) 
+
+	if ( err != OGRERR_NONE )
 		return NULL;
 	else
 		return slot;
 }
-					
+
 static void ogrEndForeignModify (EState *estate, ResultRelInfo *rinfo)
 {
 	OgrFdwModifyState *modstate = rinfo->ri_FdwState;
-	
+
 	elog(DEBUG2, "ogrEndForeignModify");
-	
+
 	ogrFinishConnection( &(modstate->ogr) );
-	
+
 	return;
 }
-				
+
 static int ogrIsForeignRelUpdatable (Relation rel)
 {
 	static int readonly = 0;
@@ -2449,7 +2484,7 @@ static int ogrIsForeignRelUpdatable (Relation rel)
 		elog(NOTICE, "no \"fid\" column in foreign table '%s'", get_rel_name(foreigntableid));
 		return readonly;
 	}
-	
+
 	/*   Is it backed by a writable OGR driver? */
 	/*   Can we open the relation in read/write mode? */
 	ogr = ogrGetConnectionFromTable(foreigntableid, true);
@@ -2461,14 +2496,14 @@ static int ogrIsForeignRelUpdatable (Relation rel)
 
 	if ( OGR_L_TestCapability(ogr.lyr, OLCSequentialWrite) )
 		updateable |= (1 << CMD_INSERT);
-	
+
 	if ( OGR_L_TestCapability(ogr.lyr, OLCDeleteFeature) )
 		updateable |= (1 << CMD_DELETE);
-	
+
 	ogrFinishConnection(&ogr);
-		
+
 	return updateable;
-	
+
 }
 
 #if PG_VERSION_NUM >= 90500
@@ -2575,15 +2610,15 @@ ogrImportForeignSchema(ImportForeignSchemaStmt *stmt, Oid serverOid)
 			OGRErr err;
 			stringbuffer_t buf;
 			stringbuffer_init(&buf);
-			
+
 			err = ogrLayerToSQL(ogr_lyr,
-			         quote_identifier(server->servername), 
-			         launder_table_names, 
+			         quote_identifier(server->servername),
+			         launder_table_names,
 			         launder_column_names,
-			         GEOMETRYOID != BYTEAOID, 
+			         GEOMETRYOID != BYTEAOID,
 			         &buf
 			      );
-					 
+
 			if (err != OGRERR_NONE)
 			{
 				elog(ERROR, "unable to generate IMPORT SQL for '%s'", table_name);
@@ -2595,7 +2630,7 @@ ogrImportForeignSchema(ImportForeignSchemaStmt *stmt, Oid serverOid)
 	}
 
 	elog(NOTICE, "Number of tables to be created %d", list_length(commands) );
-	
+
 	ogrFinishConnection(&ogr);
 
 	return commands;
